@@ -1,21 +1,21 @@
-/* Copyright (C) 1998,1999,2000,2001, 2006, 2008, 2009, 2011,
- *   2012, 2013 Free Software Foundation, Inc.
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
- * the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
- * 02110-1301 USA
- */
+/* Copyright 1998-2001,2006,2008-2009,2011-2013,2018
+     Free Software Foundation, Inc.
+
+   This file is part of Guile.
+
+   Guile is free software: you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as published
+   by the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
+
+   Guile is distributed in the hope that it will be useful, but WITHOUT
+   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+   License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with Guile.  If not, see
+   <https://www.gnu.org/licenses/>.  */
 
 
 /* This is an implementation of guardians as described in
@@ -49,20 +49,22 @@
 # include <config.h>
 #endif
 
-#include "libguile/_scm.h"
-#include "libguile/ports.h"
-#include "libguile/print.h"
-#include "libguile/smob.h"
-#include "libguile/validate.h"
-#include "libguile/root.h"
-#include "libguile/hashtab.h"
-#include "libguile/deprecation.h"
-#include "libguile/eval.h"
-
-#include "libguile/guardians.h"
-#include "libguile/bdw-gc.h"
-
+#include "bdw-gc.h"
+#include "boolean.h"
+#include "deprecation.h"
+#include "eval.h"
+#include "gsubr.h"
+#include "hashtab.h"
+#include "list.h"
+#include "numbers.h"
+#include "pairs.h"
+#include "ports.h"
+#include "print.h"
+#include "smob.h"
+#include "threads.h"
+#include "weak-vector.h"
 
+#include "guardians.h"
 
 
 static scm_t_bits tc16_guardian;
@@ -375,11 +377,5 @@ scm_init_guardians ()
   scm_set_smob_print (tc16_guardian, guardian_print);
   scm_set_smob_apply (tc16_guardian, guardian_apply, 0, 1, 0);
 
-#include "libguile/guardians.x"
+#include "guardians.x"
 }
-
-/*
-  Local Variables:
-  c-file-style: "gnu"
-  End:
-*/
