@@ -1,4 +1,4 @@
-/* Copyright 1995-2014,2016-2018
+/* Copyright 1995-2014,2016-2019
      Free Software Foundation, Inc.
 
    This file is part of Guile.
@@ -376,11 +376,11 @@ SCM_DEFINE (scm_getpwuid, "getpw", 0, 1, 0,
   SCM_SIMPLE_VECTOR_SET(result, 3, scm_from_ulong (entry->pw_gid));
   SCM_SIMPLE_VECTOR_SET(result, 4, scm_from_locale_string (entry->pw_gecos));
   if (!entry->pw_dir)
-    SCM_SIMPLE_VECTOR_SET(result, 5, scm_from_locale_string (""));
+    SCM_SIMPLE_VECTOR_SET(result, 5, scm_from_utf8_string (""));
   else
     SCM_SIMPLE_VECTOR_SET(result, 5, scm_from_locale_string (entry->pw_dir));
   if (!entry->pw_shell)
-    SCM_SIMPLE_VECTOR_SET(result, 6, scm_from_locale_string (""));
+    SCM_SIMPLE_VECTOR_SET(result, 6, scm_from_utf8_string (""));
   else
     SCM_SIMPLE_VECTOR_SET(result, 6, scm_from_locale_string (entry->pw_shell));
   return result;
@@ -1626,7 +1626,10 @@ SCM_DEFINE (scm_utime, "utime", 1, 5, 0,
 	    "(utime \"foo\" (- (current-time) 3600))\n"
 	    "@end lisp\n"
 	    "will set the access time to one hour in the past and the\n"
-	    "modification time to the current time.")
+	    "modification time to the current time.\n\n"
+            "Last, @var{flags} may be either @code{0} or the\n"
+            "@code{AT_SYMLINK_NOFOLLOW} constant, to set the time of\n"
+            "@var{pathname} even if it is a symbolic link.\n")
 #define FUNC_NAME s_scm_utime
 {
   int rv;
@@ -2435,6 +2438,19 @@ scm_init_posix ()
   scm_c_define ("LOCK_NB", scm_from_int (LOCK_NB));
 #endif
 
+#ifdef AT_SYMLINK_NOFOLLOW
+  scm_c_define ("AT_SYMLINK_NOFOLLOW", scm_from_int (AT_SYMLINK_NOFOLLOW));
+#endif
+#ifdef AT_SYMLINK_FOLLOW
+  scm_c_define ("AT_SYMLINK_FOLLOW", scm_from_int (AT_SYMLINK_FOLLOW));
+#endif
+#ifdef AT_NO_AUTOMOUNT
+  scm_c_define ("AT_NO_AUTOMOUNT", scm_from_int (AT_NO_AUTOMOUNT));
+#endif
+#ifdef AT_EMPTY_PATH
+  scm_c_define ("AT_EMPTY_PATH", scm_from_int (AT_EMPTY_PATH));
+#endif
+
 #include "cpp-SIG.c"
 #include "posix.x"
 
