@@ -1,21 +1,21 @@
-/* Copyright (C) 1995, 1996, 1998, 2000, 2001, 2004, 2006,
- *   2008-2016 Free Software Foundation, Inc.
- * 
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
- * the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
- * 02110-1301 USA
- */
+/* Copyright 1995-1996,1998,2000-2001,2004,2006,2008-2016,2018-2019
+     Free Software Foundation, Inc.
+
+   This file is part of Guile.
+
+   Guile is free software: you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as published
+   by the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
+
+   Guile is distributed in the hope that it will be useful, but WITHOUT
+   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+   License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with Guile.  If not, see
+   <https://www.gnu.org/licenses/>.  */
 
 
 
@@ -27,24 +27,32 @@
 #include <string.h>
 #include <stdio.h>
 #include <ctype.h>
+#include <errno.h>
 #include <uninorm.h>
 #include <unistr.h>
 #include <uniconv.h>
 #include <c-strcase.h>
 
+#include "chars.h"
+#include "deprecation.h"
+#include "error.h"
+#include "generalized-vectors.h"
+#include "gsubr.h"
+#include "numbers.h"
+#include "pairs.h"
+#include "ports-internal.h"
+#include "ports.h"
+#include "private-options.h"
 #include "striconveh.h"
+#include "symbols.h"
+#include "threads.h"
 
-#include "libguile/_scm.h"
-#include "libguile/chars.h"
-#include "libguile/root.h"
-#include "libguile/strings.h"
-#include "libguile/ports.h"
-#include "libguile/ports-internal.h"
-#include "libguile/error.h"
-#include "libguile/generalized-vectors.h"
-#include "libguile/deprecation.h"
-#include "libguile/validate.h"
-#include "libguile/private-options.h"
+#include "strings.h"
+
+
+#ifndef SCM_MAX_ALLOCA
+# define SCM_MAX_ALLOCA 4096 /* Max bytes per string to allocate via alloca */
+#endif
 
 
 
@@ -55,40 +63,38 @@ SCM_SYMBOL (sym_UTF_8, "UTF-8");
 SCM_SYMBOL (sym_ISO_8859_1, "ISO-8859-1");
 SCM_SYMBOL (sym_error, "error");
 
-/* Stringbufs 
- *
- * XXX - keeping an accurate refcount during GC seems to be quite
- * tricky, so we just keep score of whether a stringbuf might be
- * shared, not whether it definitely is.  
- *
- * The scheme I (mvo) tried to keep an accurate reference count would
- * recount all strings that point to a stringbuf during the mark-phase
- * of the GC.  This was done since one cannot access the stringbuf of
- * a string when that string is freed (in order to decrease the
- * reference count).  The memory of the stringbuf might have been
- * reused already for something completely different.
- *
- * This recounted worked for a small number of threads beating on
- * cow-strings, but it failed randomly with more than 10 threads, say.
- * I couldn't figure out what went wrong, so I used the conservative
- * approach implemented below.
- *
- * There are 2 storage strategies for stringbufs: 8-bit and wide.  8-bit
- * strings are ISO-8859-1-encoded strings; wide strings are 32-bit (UCS-4)
- * strings.
- */
+/* A stringbuf is a linear buffer of characters.  Every string has a
+   stringbuf.  Strings may reference just a slice of a stringbuf; that's
+   often the case for strings made by the "substring" function.
+
+   Stringbufs may hold either 8-bit characters or 32-bit characters.  In
+   either case the characters are Unicode codepoints.  "Narrow"
+   stringbufs thus have the ISO-8859-1 (Latin-1) encoding, and "wide"
+   stringbufs have the UTF-32 (UCS-4) encoding.
+
+   By default, stringbufs are immutable.  This enables an O(1)
+   "substring" operation with no synchronization.  A string-set! will
+   first ensure that the string's stringbuf is mutable, copying the
+   stringbuf if necessary.  This is therefore a copy-on-write
+   representation.  However, taking a substring of a mutable stringbuf
+   is an O(n) operation as it has to create a new immutable stringbuf.
+   There are also mutation-sharing substrings as well.  */
 
 /* The size in words of the stringbuf header (type tag + size).  */
 #define STRINGBUF_HEADER_SIZE   2U
 
 #define STRINGBUF_HEADER_BYTES  (STRINGBUF_HEADER_SIZE * sizeof (SCM))
 
-#define STRINGBUF_F_SHARED      SCM_I_STRINGBUF_F_SHARED
 #define STRINGBUF_F_WIDE        SCM_I_STRINGBUF_F_WIDE
+#define STRINGBUF_F_MUTABLE     SCM_I_STRINGBUF_F_MUTABLE
 
 #define STRINGBUF_TAG           scm_tc7_stringbuf
-#define STRINGBUF_SHARED(buf)   (SCM_CELL_WORD_0(buf) & STRINGBUF_F_SHARED)
 #define STRINGBUF_WIDE(buf)     (SCM_CELL_WORD_0(buf) & STRINGBUF_F_WIDE)
+#define STRINGBUF_MUTABLE(buf)  (SCM_CELL_WORD_0(buf) & STRINGBUF_F_MUTABLE)
+
+
+#define STRINGBUF_SET_MUTABLE(buf) \
+  SCM_SET_CELL_WORD_0 (buf, SCM_CELL_WORD_0 (buf) | STRINGBUF_F_MUTABLE)
 
 #define STRINGBUF_CONTENTS(buf) ((void *)				\
                                  SCM_CELL_OBJECT_LOC (buf,		\
@@ -98,16 +104,6 @@ SCM_SYMBOL (sym_error, "error");
 
 #define STRINGBUF_LENGTH(buf)   (SCM_CELL_WORD_1 (buf))
 
-#define SET_STRINGBUF_SHARED(buf)					\
-  do									\
-    {									\
-      /* Don't modify BUF if it's already marked as shared since it might be \
-	 a read-only, statically allocated stringbuf.  */		\
-      if (SCM_LIKELY (!STRINGBUF_SHARED (buf)))				\
-	SCM_SET_CELL_WORD_0 ((buf), SCM_CELL_WORD_0 (buf) | STRINGBUF_F_SHARED); \
-    }									\
-  while (0)
-
 #ifdef SCM_STRING_LENGTH_HISTOGRAM
 static size_t lenhist[1001];
 #endif
@@ -232,8 +228,6 @@ narrow_stringbuf (SCM buf)
   return new_buf;
 }
 
-scm_i_pthread_mutex_t stringbuf_write_mutex = SCM_I_PTHREAD_MUTEX_INITIALIZER;
-
 
 /* Copy-on-write strings.
  */
@@ -268,15 +262,8 @@ scm_i_pthread_mutex_t stringbuf_write_mutex = SCM_I_PTHREAD_MUTEX_INITIALIZER;
 void
 scm_i_print_stringbuf (SCM exp, SCM port, scm_print_state *pstate) 
 {
-  SCM str;
-
-  scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-  SET_STRINGBUF_SHARED (exp);
-  scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
-
-  str =  scm_double_cell (RO_STRING_TAG, SCM_UNPACK(exp),
-                          0, STRINGBUF_LENGTH (exp));
-
+  SCM str = scm_double_cell (STRING_TAG, SCM_UNPACK(exp), 0,
+                             STRINGBUF_LENGTH (exp));
   scm_puts ("#<stringbuf ", port);
   scm_iprin1 (str, port, pstate);
   scm_puts (">", port);
@@ -290,7 +277,6 @@ static void
 init_null_stringbuf (void)
 {
   null_stringbuf = make_stringbuf (0);
-  SET_STRINGBUF_SHARED (null_stringbuf);
 }
 
 /* Create a scheme string with space for LEN 8-bit Latin-1-encoded
@@ -360,77 +346,109 @@ get_str_buf_start (SCM *str, SCM *buf, size_t *start)
   *buf = STRING_STRINGBUF (*str);
 }
 
+static SCM
+substring_with_immutable_stringbuf (SCM str, size_t start, size_t end,
+                                    int force_copy_p, int read_only_p)
+{
+  SCM buf;
+  size_t str_start, len;
+  scm_t_bits tag = read_only_p ? RO_STRING_TAG : STRING_TAG;
+
+  get_str_buf_start (&str, &buf, &str_start);
+  len = end - start;
+  start += str_start;
+
+  if (len == 0)
+    return scm_i_make_string (0, NULL, read_only_p);
+  else if (!force_copy_p && SCM_LIKELY (!STRINGBUF_MUTABLE (buf)))
+    return scm_double_cell (tag, SCM_UNPACK (buf), start, len);
+  else
+    {
+      SCM new_buf, new_str;
+
+      if (STRINGBUF_WIDE (buf))
+        {
+          new_buf = make_wide_stringbuf (len);
+          u32_cpy ((uint32_t *) STRINGBUF_WIDE_CHARS (new_buf),
+                   (uint32_t *) (STRINGBUF_WIDE_CHARS (buf) + start), len);
+          new_str = scm_double_cell (tag, SCM_UNPACK (new_buf), 0, len);
+          scm_i_try_narrow_string (new_str);
+        }
+      else
+        {
+          new_buf = make_stringbuf (len);
+          memcpy (STRINGBUF_CHARS (new_buf),
+                  STRINGBUF_CHARS (buf) + start, len);
+          new_str = scm_double_cell (tag, SCM_UNPACK (new_buf), 0, len);
+        }
+
+      return new_str;
+    }
+}
+
 SCM
 scm_i_substring (SCM str, size_t start, size_t end)
 {
-  if (start == end)
-    return scm_i_make_string (0, NULL, 0);
-  else
-    {
-      SCM buf;
-      size_t str_start;
-      get_str_buf_start (&str, &buf, &str_start);
-      scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-      SET_STRINGBUF_SHARED (buf);
-      scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
-      return scm_double_cell (STRING_TAG, SCM_UNPACK(buf),
-                              (scm_t_bits)str_start + start,
-                              (scm_t_bits) end - start);
-    }
+  return substring_with_immutable_stringbuf (str, start, end, 0, 0);
 }
 
 SCM
 scm_i_substring_read_only (SCM str, size_t start, size_t end)
 {
-  if (start == end)
-    return scm_i_make_string (0, NULL, 1);
-  else
-    {
-      SCM buf;
-      size_t str_start;
-      get_str_buf_start (&str, &buf, &str_start);
-      scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-      SET_STRINGBUF_SHARED (buf);
-      scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
-      return scm_double_cell (RO_STRING_TAG, SCM_UNPACK(buf),
-                              (scm_t_bits)str_start + start,
-                              (scm_t_bits) end - start);
-    }
+  return substring_with_immutable_stringbuf (str, start, end, 0, 1);
 }
 
 SCM
 scm_i_substring_copy (SCM str, size_t start, size_t end)
 {
-  if (start == end)
-    return scm_i_make_string (0, NULL, 0);
-  else
+  return substring_with_immutable_stringbuf (str, start, end, 1, 0);
+}
+
+static void
+scm_i_string_ensure_mutable_x (SCM str)
+{
+  SCM buf;
+
+  if (IS_SH_STRING (str))
     {
-      size_t len = end - start;
-      SCM buf, my_buf, substr;
-      size_t str_start;
-      int wide = 0;
-      get_str_buf_start (&str, &buf, &str_start);
-      if (scm_i_is_narrow_string (str))
-        {
-          my_buf = make_stringbuf (len);
-          memcpy (STRINGBUF_CHARS (my_buf),
-                  STRINGBUF_CHARS (buf) + str_start + start, len);
-        }
-      else
-        {
-          my_buf = make_wide_stringbuf (len);
-          u32_cpy ((scm_t_uint32 *) STRINGBUF_WIDE_CHARS (my_buf),
-                   (scm_t_uint32 *) (STRINGBUF_WIDE_CHARS (buf) + str_start 
-                                     + start), len);
-          wide = 1;
-        }
-      scm_remember_upto_here_1 (buf);
-      substr = scm_double_cell (STRING_TAG, SCM_UNPACK (my_buf),
-                                (scm_t_bits) 0, (scm_t_bits) len);
-      if (wide)
-        scm_i_try_narrow_string (substr);
-      return substr;
+      /* Shared-mutation strings always have mutable stringbufs.  */
+      buf = STRING_STRINGBUF (SH_STRING_STRING (str));
+      if (!STRINGBUF_MUTABLE (buf))
+        abort ();
+      return;
     }
+
+  if (IS_RO_STRING (str))
+    scm_misc_error (NULL, "string is read-only: ~s", scm_list_1 (str));
+
+  buf = STRING_STRINGBUF (str);
+
+  if (STRINGBUF_MUTABLE (buf))
+    return;
+
+  /* Otherwise copy and mark the fresh stringbuf as mutable.  Note that
+     we copy the whole stringbuf so that the start/len offsets from the
+     original string keep working, so that concurrent accessors on this
+     string don't see things in an inconsistent state.  */
+  {
+    SCM new_buf;
+    size_t len = STRINGBUF_LENGTH (buf);
+
+    if (STRINGBUF_WIDE (buf))
+      {
+        new_buf = make_wide_stringbuf (len);
+        u32_cpy ((uint32_t *) STRINGBUF_WIDE_CHARS (new_buf),
+                 (uint32_t *) STRINGBUF_WIDE_CHARS (buf), len);
+      }
+    else
+      {
+        new_buf = make_stringbuf (len);
+        memcpy (STRINGBUF_CHARS (new_buf), STRINGBUF_CHARS (buf), len);
+      }
+
+    STRINGBUF_SET_MUTABLE (new_buf);
+    SET_STRING_STRINGBUF (str, new_buf);
+  }
 }
 
 SCM
@@ -440,6 +458,8 @@ scm_i_substring_shared (SCM str, size_t start, size_t end)
     return str;
   else if (start == end)
     return scm_i_make_string (0, NULL, 0);
+  else if (IS_RO_STRING (str))
+    return scm_i_substring_read_only (str, start, end);
   else
     {
       size_t len = end - start;
@@ -448,6 +468,9 @@ scm_i_substring_shared (SCM str, size_t start, size_t end)
 	  start += STRING_START (str);
 	  str = SH_STRING_STRING (str);
 	}
+
+      scm_i_string_ensure_mutable_x (str);
+
       return scm_double_cell (SH_STRING_TAG, SCM_UNPACK(str),
 			      (scm_t_bits)start, (scm_t_bits) len);
     }
@@ -493,6 +516,12 @@ scm_i_string_length (SCM str)
   return STRING_LENGTH (str);
 }
 
+int
+scm_i_string_is_mutable (SCM str)
+{
+  return !IS_RO_STRING (str);
+}
+
 /* True if the string is 'narrow', meaning it has a 8-bit Latin-1
    encoding.  False if it is 'wide', having a 32-bit UCS-4
    encoding.  */
@@ -569,60 +598,13 @@ scm_i_string_wide_chars (SCM str)
                     scm_list_1 (str));
 }
 
-/* If the buffer in ORIG_STR is shared, copy ORIG_STR's characters to
+/* If the buffer in ORIG_STR is immutable, copy ORIG_STR's characters to
    a new string buffer, so that it can be modified without modifying
-   other strings.  Also, lock the string mutex.  Later, one must call
-   scm_i_string_stop_writing to unlock the mutex.  */
+   other strings.  */
 SCM
 scm_i_string_start_writing (SCM orig_str)
 {
-  SCM buf, str = orig_str;
-  size_t start;
-
-  get_str_buf_start (&str, &buf, &start);
-  if (IS_RO_STRING (str))
-    scm_misc_error (NULL, "string is read-only: ~s", scm_list_1 (orig_str));
-
-  scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-  if (STRINGBUF_SHARED (buf))
-    {
-      /* Clone the stringbuf.  */
-      size_t len = STRING_LENGTH (str);
-      SCM new_buf;
-
-      scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
-
-      if (scm_i_is_narrow_string (str))
-        {
-          new_buf = make_stringbuf (len);
-          memcpy (STRINGBUF_CHARS (new_buf),
-                  STRINGBUF_CHARS (buf) + STRING_START (str), len);
-
-        }
-      else
-        {
-          new_buf = make_wide_stringbuf (len);
-          u32_cpy ((scm_t_uint32 *) STRINGBUF_WIDE_CHARS (new_buf),
-                   (scm_t_uint32 *) (STRINGBUF_WIDE_CHARS (buf) 
-                                     + STRING_START (str)), len);
-        }
-
-      SET_STRING_STRINGBUF (str, new_buf);
-      start -= STRING_START (str);
-
-      /* FIXME: The following operations are not atomic, so other threads
-	 looking at STR may see an inconsistent state.  Nevertheless it can't
-	 hurt much since (i) accessing STR while it is being mutated can't
-	 yield a crash, and (ii) concurrent accesses to STR should be
-	 protected by a mutex at the application level.  The latter may not
-	 apply when STR != ORIG_STR, though.  */
-      SET_STRING_START (str, 0);
-      SET_STRING_STRINGBUF (str, new_buf);
-
-      buf = new_buf;
-
-      scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-    }
+  scm_i_string_ensure_mutable_x (orig_str);
   return orig_str;
 }
 
@@ -662,7 +644,6 @@ scm_i_string_writable_wide_chars (SCM str)
 void
 scm_i_string_stop_writing (void)
 {
-  scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
 }
 
 /* Return the Xth character of STR as a UCS-4 codepoint.  */
@@ -769,42 +750,10 @@ scm_i_make_symbol (SCM name, scm_t_bits flags,
 		   unsigned long hash, SCM props)
 {
   SCM buf;
-  size_t start = STRING_START (name);
   size_t length = STRING_LENGTH (name);
 
-  if (IS_SH_STRING (name))
-    {
-      name = SH_STRING_STRING (name);
-      start += STRING_START (name);
-    }
+  name = scm_i_substring_copy (name, 0, length);
   buf = STRING_STRINGBUF (name);
-
-  if (start == 0 && length == STRINGBUF_LENGTH (buf))
-    {
-      /* reuse buf. */
-      scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-      SET_STRINGBUF_SHARED (buf);
-      scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
-    }
-  else
-    {
-      /* make new buf. */
-      if (scm_i_is_narrow_string (name))
-        {
-          SCM new_buf = make_stringbuf (length);
-          memcpy (STRINGBUF_CHARS (new_buf),
-                  STRINGBUF_CHARS (buf) + start, length);
-          buf = new_buf;
-        }
-      else
-        {
-          SCM new_buf = make_wide_stringbuf (length);
-          u32_cpy ((scm_t_uint32 *) STRINGBUF_WIDE_CHARS (new_buf),
-                   (scm_t_uint32 *) STRINGBUF_WIDE_CHARS (buf) + start,
-                   length);
-          buf = new_buf;
-        }
-    }
   return scm_double_cell (scm_tc7_symbol | flags, SCM_UNPACK (buf),
 			  (scm_t_bits) hash, SCM_UNPACK (props));
 }
@@ -883,9 +832,6 @@ SCM
 scm_i_symbol_substring (SCM sym, size_t start, size_t end)
 {
   SCM buf = SYMBOL_STRINGBUF (sym);
-  scm_i_pthread_mutex_lock (&stringbuf_write_mutex);
-  SET_STRINGBUF_SHARED (buf);
-  scm_i_pthread_mutex_unlock (&stringbuf_write_mutex);
   return scm_double_cell (RO_STRING_TAG, SCM_UNPACK (buf),
 			  (scm_t_bits)start, (scm_t_bits) end - start);
 }
@@ -922,8 +868,8 @@ SCM_DEFINE (scm_sys_string_dump, "%string-dump", 1, 0, 0, (SCM str),
             "A new string containing this string's stringbuf's characters\n"
             "@item stringbuf-length\n"
             "The number of characters in this stringbuf\n"
-            "@item stringbuf-shared\n"
-            "@code{#t} if this stringbuf is shared\n"
+            "@item stringbuf-mutable\n"
+            "@code{#t} if this stringbuf is mutable\n"
             "@item stringbuf-wide\n"
             "@code{#t} if this stringbuf's characters are stored in a\n"
             "32-bit buffer, or @code{#f} if they are stored in an 8-bit\n"
@@ -978,18 +924,18 @@ SCM_DEFINE (scm_sys_string_dump, "%string-dump", 1, 0, 0, (SCM str),
       size_t len = STRINGBUF_LENGTH (buf);
       scm_t_wchar *cbuf;
       SCM sbc = scm_i_make_wide_string (len, &cbuf, 0);
-      u32_cpy ((scm_t_uint32 *) cbuf, 
-               (scm_t_uint32 *) STRINGBUF_WIDE_CHARS (buf), len);
+      u32_cpy ((uint32_t *) cbuf, 
+               (uint32_t *) STRINGBUF_WIDE_CHARS (buf), len);
       e6 = scm_cons (scm_from_latin1_symbol ("stringbuf-chars"),
                      sbc);
     }
   e7 = scm_cons (scm_from_latin1_symbol ("stringbuf-length"), 
                  scm_from_size_t (STRINGBUF_LENGTH (buf)));
-  if (STRINGBUF_SHARED (buf))
-    e8 = scm_cons (scm_from_latin1_symbol ("stringbuf-shared"), 
+  if (STRINGBUF_MUTABLE (buf))
+    e8 = scm_cons (scm_from_latin1_symbol ("stringbuf-mutable"),
                    SCM_BOOL_T);
   else
-    e8 = scm_cons (scm_from_latin1_symbol ("stringbuf-shared"), 
+    e8 = scm_cons (scm_from_latin1_symbol ("stringbuf-mutable"),
                    SCM_BOOL_F);
   if (STRINGBUF_WIDE (buf))
     e9 = scm_cons (scm_from_latin1_symbol ("stringbuf-wide"),
@@ -1016,8 +962,8 @@ SCM_DEFINE (scm_sys_symbol_dump, "%symbol-dump", 1, 0, 0, (SCM sym),
             "A new string containing this symbols's stringbuf's characters\n"
             "@item stringbuf-length\n"
             "The number of characters in this stringbuf\n"
-            "@item stringbuf-shared\n"
-            "@code{#t} if this stringbuf is shared\n"
+            "@item stringbuf-mutable\n"
+            "@code{#t} if this stringbuf is mutable\n"
             "@item stringbuf-wide\n"
             "@code{#t} if this stringbuf's characters are stored in a\n"
             "32-bit buffer, or @code{#f} if they are stored in an 8-bit\n"
@@ -1051,18 +997,18 @@ SCM_DEFINE (scm_sys_symbol_dump, "%symbol-dump", 1, 0, 0, (SCM sym),
       size_t len = STRINGBUF_LENGTH (buf);
       scm_t_wchar *cbuf;
       SCM sbc = scm_i_make_wide_string (len, &cbuf, 0);
-      u32_cpy ((scm_t_uint32 *) cbuf, 
-               (scm_t_uint32 *) STRINGBUF_WIDE_CHARS (buf), len);
+      u32_cpy ((uint32_t *) cbuf, 
+               (uint32_t *) STRINGBUF_WIDE_CHARS (buf), len);
       e4 = scm_cons (scm_from_latin1_symbol ("stringbuf-chars"),
                      sbc);
     }
   e5 = scm_cons (scm_from_latin1_symbol ("stringbuf-length"), 
                  scm_from_size_t (STRINGBUF_LENGTH (buf)));
-  if (STRINGBUF_SHARED (buf))
-    e6 = scm_cons (scm_from_latin1_symbol ("stringbuf-shared"), 
+  if (STRINGBUF_MUTABLE (buf))
+    e6 = scm_cons (scm_from_latin1_symbol ("stringbuf-mutable"),
                    SCM_BOOL_T);
   else
-    e6 = scm_cons (scm_from_latin1_symbol ("stringbuf-shared"), 
+    e6 = scm_cons (scm_from_latin1_symbol ("stringbuf-mutable"),
                    SCM_BOOL_F);
   if (STRINGBUF_WIDE (buf))
     e7 = scm_cons (scm_from_latin1_symbol ("stringbuf-wide"),
@@ -1146,7 +1092,6 @@ SCM_DEFINE (scm_string, "string", 0, 0, 1,
       char *buf;
 
       result = scm_i_make_string (len, NULL, 0);
-      result = scm_i_string_start_writing (result);
       buf = scm_i_string_writable_chars (result);
       while (len > 0 && scm_is_pair (rest))
         {
@@ -1163,7 +1108,6 @@ SCM_DEFINE (scm_string, "string", 0, 0, 1,
       scm_t_wchar *buf;
 
       result = scm_i_make_wide_string (len, NULL, 0);
-      result = scm_i_string_start_writing (result);
       buf = scm_i_string_writable_wide_chars (result);
       while (len > 0 && scm_is_pair (rest))
         {
@@ -1175,7 +1119,6 @@ SCM_DEFINE (scm_string, "string", 0, 0, 1,
           scm_remember_upto_here_1 (elt);
         }
     }
-  scm_i_string_stop_writing ();
 
   if (len > 0)
     scm_misc_error (NULL, "list changed while constructing string", SCM_EOL);
@@ -1194,7 +1137,12 @@ SCM_DEFINE (scm_make_string, "make-string", 1, 1, 0,
 	    "of the string are all set to @code{#\nul}.")
 #define FUNC_NAME s_scm_make_string
 {
-  return scm_c_make_string (scm_to_size_t (k), chr);
+  SCM ret = scm_c_make_string (scm_to_size_t (k), chr);
+  /* Given that make-string is mostly used by Scheme to prepare a
+     mutable string buffer, let's go ahead and mark this as mutable to
+     avoid a copy when this buffer is next written to.  */
+  STRINGBUF_SET_MUTABLE (STRING_STRINGBUF (ret));
+  return ret;
 }
 #undef FUNC_NAME
 
@@ -1212,10 +1160,8 @@ scm_c_make_string (size_t len, SCM chr)
   else
     {
       SCM_VALIDATE_CHAR (0, chr);
-      res = scm_i_string_start_writing (res);
       for (p = 0; p < len; p++)
         scm_i_string_set_x (res, p, SCM_CHAR (chr));
-      scm_i_string_stop_writing ();
     }
 
   return res;
@@ -1272,9 +1218,9 @@ SCM_DEFINE (scm_string_ref, "string-ref", 2, 0, 0,
     scm_out_of_range (NULL, k);
 
   if (scm_i_is_narrow_string (str))
-    return SCM_MAKE_CHAR (scm_i_string_chars (str)[idx]);
+    return scm_c_make_char (scm_i_string_chars (str)[idx]);
   else
-    return SCM_MAKE_CHAR (scm_i_string_wide_chars (str)[idx]);
+    return scm_c_make_char (scm_i_string_wide_chars (str)[idx]);
 }
 #undef FUNC_NAME
 
@@ -1284,9 +1230,9 @@ scm_c_string_ref (SCM str, size_t p)
   if (p >= scm_i_string_length (str))
     scm_out_of_range (NULL, scm_from_size_t (p));
   if (scm_i_is_narrow_string (str))
-    return SCM_MAKE_CHAR (scm_i_string_chars (str)[p]);
+    return scm_c_make_char (scm_i_string_chars (str)[p]);
   else
-    return SCM_MAKE_CHAR (scm_i_string_wide_chars (str)[p]);
+    return scm_c_make_char (scm_i_string_wide_chars (str)[p]);
 
 }
 
@@ -1481,8 +1427,8 @@ SCM_DEFINE (scm_string_append, "string-append", 0, 0, 1,
                 data.wide[i] = (unsigned char) src[i];
             }
           else
-            u32_cpy ((scm_t_uint32 *) data.wide,
-                     (scm_t_uint32 *) scm_i_string_wide_chars (s), len);
+            u32_cpy ((uint32_t *) data.wide,
+                     (uint32_t *) scm_i_string_wide_chars (s), len);
           data.wide += len;
         }
       total -= len;
@@ -1600,7 +1546,7 @@ scm_from_stringn (const char *str, size_t len, const char *encoding,
     {
       scm_t_wchar *wdst;
       res = scm_i_make_wide_string (u32len, &wdst, 0);
-      u32_cpy ((scm_t_uint32 *) wdst, (scm_t_uint32 *) u32, u32len);
+      u32_cpy ((uint32_t *) wdst, (uint32_t *) u32, u32len);
       wdst[u32len] = 0;
     }
 
@@ -1660,7 +1606,7 @@ SCM
 scm_from_utf8_stringn (const char *str, size_t len)
 {
   size_t i, char_len;
-  const scm_t_uint8 *ustr = (const scm_t_uint8 *) str;
+  const uint8_t *ustr = (const uint8_t *) str;
   int ascii = 1, narrow = 1;
   SCM res;
 
@@ -1871,6 +1817,7 @@ static void
 unistring_escapes_to_r6rs_escapes (char *buf, size_t *lenp)
 {
   char *before, *after;
+  int malloc_p;
   size_t i, j;
   /* The worst case is if the input string contains all 4-digit hex escapes.
      "\uXXXX" (six characters) becomes "\xXXXX;" (seven characters) */
@@ -1878,7 +1825,8 @@ unistring_escapes_to_r6rs_escapes (char *buf, size_t *lenp)
   size_t nzeros, ndigits;
 
   before = buf;
-  after = alloca (max_out_len);
+  malloc_p = (max_out_len > SCM_MAX_ALLOCA);
+  after = malloc_p ? malloc (max_out_len) : alloca (max_out_len);
   i = 0;
   j = 0;
   while (i < *lenp)
@@ -1936,6 +1884,8 @@ unistring_escapes_to_r6rs_escapes (char *buf, size_t *lenp)
     }
   *lenp = j;
   memcpy (before, after, j);
+  if (malloc_p)
+    free (after);
 }
 
 char *
@@ -1990,7 +1940,7 @@ scm_to_utf8_string (SCM str)
 }
 
 static size_t
-latin1_u8_strlen (const scm_t_uint8 *str, size_t len)
+latin1_u8_strlen (const uint8_t *str, size_t len)
 {
   size_t ret, i;
   for (i = 0, ret = 0; i < len; i++)
@@ -1998,9 +1948,9 @@ latin1_u8_strlen (const scm_t_uint8 *str, size_t len)
   return ret;
 }
 
-static scm_t_uint8*
-latin1_to_u8 (const scm_t_uint8 *str, size_t latin_len,
-              scm_t_uint8 *u8_result, size_t *u8_lenp)
+static uint8_t*
+latin1_to_u8 (const uint8_t *str, size_t latin_len,
+              uint8_t *u8_result, size_t *u8_lenp)
 {
   size_t i, n;
   size_t u8_len = latin1_u8_strlen (str, latin_len);
@@ -2040,13 +1990,13 @@ latin1_to_u8 (const scm_t_uint8 *str, size_t latin_len,
 */
 
 static size_t
-u32_u8_length_in_bytes (const scm_t_uint32 *str, size_t len)
+u32_u8_length_in_bytes (const uint32_t *str, size_t len)
 {
   size_t ret, i;
 
   for (i = 0, ret = 0; i < len; i++)
     {
-      scm_t_uint32 c = str[i];
+      uint32_t c = str[i];
 
       if (c <= 0x7f)
         ret += 1;
@@ -2069,11 +2019,11 @@ static size_t
 utf8_length (SCM str)
 {
   if (scm_i_is_narrow_string (str))
-    return latin1_u8_strlen ((scm_t_uint8 *) scm_i_string_chars (str),
+    return latin1_u8_strlen ((uint8_t *) scm_i_string_chars (str),
                              scm_i_string_length (str));
   else
     return u32_u8_length_in_bytes
-      ((scm_t_uint32 *) scm_i_string_wide_chars (str),
+      ((uint32_t *) scm_i_string_wide_chars (str),
        scm_i_string_length (str));
 }
 
@@ -2104,13 +2054,13 @@ scm_to_utf8_stringn (SCM str, size_t *lenp)
   SCM_VALIDATE_STRING (1, str);
 
   if (scm_i_is_narrow_string (str))
-    return (char *) latin1_to_u8 ((scm_t_uint8 *) scm_i_string_chars (str),
+    return (char *) latin1_to_u8 ((uint8_t *) scm_i_string_chars (str),
                                   scm_i_string_length (str),
                                   NULL, lenp);
   else
     {
-      scm_t_uint32 *chars = (scm_t_uint32 *) scm_i_string_wide_chars (str);
-      scm_t_uint8 *buf, *ret;
+      uint32_t *chars = (uint32_t *) scm_i_string_wide_chars (str);
+      uint8_t *buf, *ret;
       size_t num_chars = scm_i_string_length (str);
       size_t num_bytes_predicted, num_bytes_actual;
 
@@ -2167,10 +2117,10 @@ scm_to_utf32_stringn (SCM str, size_t *lenp)
 
   if (scm_i_is_narrow_string (str))
     {
-      scm_t_uint8 *codepoints;
+      uint8_t *codepoints;
       size_t i, len;
 
-      codepoints = (scm_t_uint8*) scm_i_string_chars (str);
+      codepoints = (uint8_t*) scm_i_string_chars (str);
       len = scm_i_string_length (str);
       if (lenp)
 	*lenp = len;
@@ -2235,6 +2185,12 @@ scm_to_stringn (SCM str, size_t *lenp, const char *encoding,
 
   if (!scm_is_string (str))
     scm_wrong_type_arg_msg (NULL, 0, str, "string");
+
+  if (c_strcasecmp (encoding, "UTF-8") == 0)
+    /* This is the most common case--e.g., when calling libc bindings
+       while using a UTF-8 locale.  */
+    return scm_to_utf8_stringn (str, lenp);
+
   ilen = scm_i_string_length (str);
 
   if (ilen == 0)
@@ -2298,7 +2254,7 @@ scm_to_stringn (SCM str, size_t *lenp, const char *encoding,
     {
       buf = u32_conv_to_encoding (enc,
                                   (enum iconv_ilseq_handler) handler,
-                                  (scm_t_uint32 *) scm_i_string_wide_chars (str),
+                                  (uint32_t *) scm_i_string_wide_chars (str),
                                   ilen,
                                   NULL,
                                   NULL, &len);
@@ -2340,13 +2296,18 @@ scm_to_stringn (SCM str, size_t *lenp, const char *encoding,
 size_t
 scm_to_locale_stringbuf (SCM str, char *buf, size_t max_len)
 {
-  size_t len;
+  size_t len, copy_len;
   char *result = NULL;
   if (!scm_is_string (str))
     scm_wrong_type_arg_msg (NULL, 0, str, "string");
   result = scm_to_locale_stringn (str, &len);
 
-  memcpy (buf, result, (len > max_len) ? max_len : len);
+  copy_len = (len > max_len) ? max_len : len;
+  if (copy_len != 0)
+    /* Some users of 'scm_to_locale_stringbuf' may pass NULL for buf
+       when max_len is zero, and yet we must avoid passing NULL to
+       memcpy to avoid undefined behavior. */
+    memcpy (buf, result, copy_len);
   free (result);
 
   scm_remember_upto_here_1 (str);
@@ -2364,29 +2325,38 @@ static SCM
 normalize_str (SCM string, uninorm_t form)
 {
   SCM ret;
-  scm_t_uint32 *w_str;
+  uint32_t *w_str;
+  uint32_t *w_norm_str;
   scm_t_wchar *cbuf;
-  size_t rlen, len = scm_i_string_length (string);
+  int malloc_p;
+  size_t norm_len, len = scm_i_string_length (string);
   
   if (scm_i_is_narrow_string (string))
     {
-      size_t i;
+      size_t i, bytes;
       const char *buf = scm_i_string_chars (string);
-      
-      w_str = alloca (sizeof (scm_t_wchar) * (len + 1));
-      
+
+      bytes = (len + 1) * sizeof (scm_t_wchar);
+      malloc_p = (bytes > SCM_MAX_ALLOCA);
+      w_str = malloc_p ? malloc (bytes) : alloca (bytes);
+
       for (i = 0; i < len; i ++)
 	w_str[i] = (unsigned char) buf[i];
       w_str[len] = 0;
     }
-  else 
-    w_str = (scm_t_uint32 *) scm_i_string_wide_chars (string);
+  else
+    {
+      malloc_p = 0;
+      w_str = (uint32_t *) scm_i_string_wide_chars (string);
+    }
 
-  w_str = u32_normalize (form, w_str, len, NULL, &rlen);  
-  
-  ret = scm_i_make_wide_string (rlen, &cbuf, 0);
-  u32_cpy ((scm_t_uint32 *) cbuf, w_str, rlen);
-  free (w_str);
+  w_norm_str = u32_normalize (form, w_str, len, NULL, &norm_len);
+
+  ret = scm_i_make_wide_string (norm_len, &cbuf, 0);
+  u32_cpy ((uint32_t *) cbuf, w_norm_str, norm_len);
+  free (w_norm_str);
+  if (malloc_p)
+    free (w_str);
 
   scm_i_try_narrow_string (ret);
 
@@ -2517,12 +2487,6 @@ scm_init_strings ()
 {
   scm_nullstr = scm_i_make_string (0, NULL, 0);
 
-#include "libguile/strings.x"
+#include "strings.x"
 }
 
-
-/*
-  Local Variables:
-  c-file-style: "gnu"
-  End:
-*/
