@@ -1,35 +1,33 @@
-/* classes: h_files */
-
 #ifndef SCM_THREADS_H
 #define SCM_THREADS_H
 
-/* Copyright (C) 1996, 1997, 1998, 2000, 2001, 2002, 2003, 2004, 2006,
- *   2007, 2008, 2009, 2011, 2012, 2013, 2014 Free Software Foundation, Inc.
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation; either version 3 of
- * the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
- * 02110-1301 USA
- */
+/* Copyright 1996-1998,2000-2004,2006-2009,2011-2014,2018
+     Free Software Foundation, Inc.
+
+   This file is part of Guile.
+
+   Guile is free software: you can redistribute it and/or modify it
+   under the terms of the GNU Lesser General Public License as published
+   by the Free Software Foundation, either version 3 of the License, or
+   (at your option) any later version.
+
+   Guile is distributed in the hope that it will be useful, but WITHOUT
+   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+   License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with Guile.  If not, see
+   <https://www.gnu.org/licenses/>.  */
 
 
 
-#include "libguile/__scm.h"
 #include "libguile/procs.h"
 #include "libguile/throw.h"
 #include "libguile/dynstack.h"
 #include "libguile/iselect.h"
-#include "libguile/continuations.h"
+#include "libguile/smob.h"
+#include "libguile/vm.h"
 
 #if SCM_USE_PTHREAD_THREADS
 #include "libguile/pthread-threads.h"
@@ -48,8 +46,11 @@ SCM_API scm_t_bits scm_tc16_condvar;
 
 struct scm_thread_wake_data;
 
-typedef struct scm_i_thread {
-  struct scm_i_thread *next_thread;
+struct scm_thread {
+  struct scm_thread *next_thread;
+
+  /* VM state for this thread.  */
+  struct scm_vm vm;
 
   SCM handle;
   scm_i_pthread_t pthread;
@@ -59,6 +60,9 @@ typedef struct scm_i_thread {
 
   /* Boolean indicating whether the thread is in guile mode.  */
   int guile_mode;
+  /* Boolean indicating whether to call GC_unregister_my_thread () when
+     this thread exits.  */
+  int needs_unregister;
 
   struct scm_thread_wake_data *wake;
   scm_i_pthread_cond_t sleep_cond;
@@ -98,18 +102,15 @@ typedef struct scm_i_thread {
   SCM continuation_root;
   SCM_STACKITEM *continuation_base;
 
-  /* For keeping track of the stack and registers. */
-  struct scm_vm *vp;
+  /* Stack base.  Used when checking for C stack overflow.  */
   SCM_STACKITEM *base;
-  scm_i_jmp_buf regs;
-#ifdef __ia64__
-  void *register_backing_store_base;
-  scm_t_contregs *pending_rbs_continuation;
-#endif
-} scm_i_thread;
+
+  /* JIT state; NULL until this thread needs to JIT-compile something.  */
+  struct scm_jit_state *jit_state;
+};
 
 #define SCM_I_IS_THREAD(x)    SCM_SMOB_PREDICATE (scm_tc16_thread, x)
-#define SCM_I_THREAD_DATA(x)  ((scm_i_thread *) SCM_SMOB_DATA (x))
+#define SCM_I_THREAD_DATA(x)  ((scm_thread *) SCM_SMOB_DATA (x))
 
 #define SCM_VALIDATE_THREAD(pos, a) \
   scm_assert_smob_type (scm_tc16_thread, (a))
@@ -172,13 +173,13 @@ SCM_INTERNAL scm_i_pthread_key_t scm_i_thread_key;
 
 # ifdef SCM_HAVE_THREAD_STORAGE_CLASS
 
-SCM_INTERNAL SCM_THREAD_LOCAL scm_i_thread *scm_i_current_thread;
+SCM_INTERNAL SCM_THREAD_LOCAL scm_thread *scm_i_current_thread;
 #  define SCM_I_CURRENT_THREAD (scm_i_current_thread)
 
 # else /* !SCM_HAVE_THREAD_STORAGE_CLASS */
 
 #  define SCM_I_CURRENT_THREAD						\
-    ((scm_i_thread *) scm_i_pthread_getspecific (scm_i_thread_key))
+    ((scm_thread *) scm_i_pthread_getspecific (scm_i_thread_key))
 
 # endif /* !SCM_HAVE_THREAD_STORAGE_CLASS */
 
@@ -210,9 +211,3 @@ SCM_API SCM scm_total_processor_count (void);
 SCM_API SCM scm_current_processor_count (void);
 
 #endif  /* SCM_THREADS_H */
-
-/*
-  Local Variables:
-  c-file-style: "gnu"
-  End:
-*/
