@@ -1,4 +1,4 @@
-/* Copyright 2001,2009-2015,2017-2018
+/* Copyright 2001,2009-2015,2017-2019
      Free Software Foundation, Inc.
 
    This file is part of Guile.
@@ -484,7 +484,9 @@ define_vm_builtins (void)
   {                                                                     \
     size_t sz = sizeof (builtin##_code);                                \
     vm_builtin_##builtin##_code = instrumented_code (builtin##_code, sz); \
-    vm_builtin_##builtin = scm_i_make_program (vm_builtin_##builtin##_code); \
+    vm_builtin_##builtin =                                              \
+      scm_cell (scm_tc7_program | SCM_F_PROGRAM_IS_PRIMITIVE,           \
+                (scm_t_bits)vm_builtin_##builtin##_code);               \
   }
   FOR_EACH_VM_BUILTIN (DEFINE_BUILTIN);
 #undef INDEX_TO_NAME
@@ -658,7 +660,10 @@ return_unused_stack_to_os (struct scm_vm *vp)
         ret = madvise ((void *) lo, hi - lo, MADV_DONTNEED);
       while (ret && errno == EAGAIN);
 
-      if (ret)
+      /* If the OS doesn't implement 'madvise' (as is currently the case
+         for GNU/Hurd), don't warn the user since there's nothing they
+         can do about it.  */
+      if (ret && errno != ENOSYS)
         perror ("madvise failed");
     }
 
