@@ -1,6 +1,6 @@
 ;;; Continuation-passing style (CPS) intermediate language (IL)
 
-;; Copyright (C) 2013, 2014, 2015 Free Software Foundation, Inc.
+;; Copyright (C) 2013, 2014, 2015, 2017, 2018 Free Software Foundation, Inc.
 
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -37,6 +37,7 @@
   #:use-module (language cps utils)
   #:use-module (language cps intmap)
   #:use-module (language cps intset)
+  #:use-module (language cps with-cps)
   #:export (contify))
 
 (define (compute-singly-referenced-labels conts)
@@ -59,8 +60,10 @@ predecessor."
       (($ $kfun src meta self ktail kclause) (ref2 ktail kclause))
       (($ $ktail) (ref0))
       (($ $kclause arity kbody kalt) (ref2 kbody kalt))
-      (($ $kargs names syms ($ $continue k src exp))
-       (ref2 k (match exp (($ $branch k) k) (($ $prompt _ _ k) k) (_ #f))))))
+      (($ $kargs names syms ($ $continue k)) (ref1 k))
+      (($ $kargs names syms ($ $branch kf kt)) (ref2 kf kt))
+      (($ $kargs names syms ($ $prompt k kh)) (ref2 k kh))
+      (($ $kargs names syms ($ $throw)) (ref0))))
   (let*-values (((single multiple) (values empty-intset empty-intset))
                 ((single multiple) (intmap-fold add-ref conts single multiple)))
     (intset-subtract (persistent-intset single)
@@ -98,24 +101,6 @@ the set."
                  conts
                  empty-intmap)))
 
-(define (compute-multi-clause conts)
-  "Compute an set containing all labels that are part of a multi-clause
-case-lambda.  See the note in compute-contification-candidates."
-  (define (multi-clause? clause)
-    (and clause
-         (match (intmap-ref conts clause)
-           (($ $kclause arity body alt)
-            alt))))
-  (intmap-fold (lambda (label cont multi)
-                 (match cont
-                   (($ $kfun src meta self tail clause)
-                    (if (multi-clause? clause)
-                        (intset-union multi (compute-function-body conts label))
-                        multi))
-                   (_ multi)))
-               conts
-               empty-intset))
-
 (define (compute-arities conts functions)
   "Given the map FUNCTIONS whose keys are $kfun labels, return a map
 from label to arities."
@@ -152,7 +137,6 @@ from label to arities."
 functions with known uses that are only ever used as the operator of a
 $call, and are always called with a compatible arity."
   (let* ((functions (compute-functions conts))
-         (multi-clause (compute-multi-clause conts))
          (vars (intmap-fold (lambda (label vars out)
                               (intset-fold (lambda (var out)
                                              (intmap-add out var label))
@@ -185,39 +169,34 @@ $call, and are always called with a compatible arity."
       (match cont
         (($ $kargs _ _ ($ $continue _ _ exp))
          (match exp
-           ((or ($ $const) ($ $prim) ($ $closure) ($ $fun) ($ $rec))
+           ((or ($ $const) ($ $prim) ($ $const-fun) ($ $code) ($ $fun) ($ $rec))
             functions)
            (($ $values args)
             (exclude-vars functions args))
            (($ $call proc args)
             (let ((functions (exclude-vars functions args)))
-              ;; This contification algorithm is happy to contify the
-              ;; `lp' in this example into a shared tail between clauses:
+              ;; Note that this contification algorithm is happy to
+              ;; contify the `lp' in this example into a shared tail
+              ;; between clauses:
               ;;
               ;; (letrec ((lp (lambda () (lp))))
               ;;   (case-lambda
               ;;     ((a) (lp))
               ;;     ((a b) (lp))))
               ;;
-              ;; However because the current compilation pipeline has to
-              ;; re-nest continuations into old CPS, there would be no
-              ;; scope in which the tail would be valid.  So, until the
-              ;; old compilation pipeline is completely replaced,
-              ;; conservatively exclude contifiable fucntions called
-              ;; from multi-clause procedures.
-              (if (intset-ref multi-clause label)
-                  (exclude-var functions proc)
-                  (restrict-arity functions proc (length args)))))
+              ;; This can cause cross-clause jumps.  The rest of the
+              ;; compiler handles this fine though, so we allow it.
+              (restrict-arity functions proc (length args))))
            (($ $callk k proc args)
             (exclude-vars functions (cons proc args)))
-           (($ $branch kt ($ $primcall name args))
-            (exclude-vars functions args))
-           (($ $branch kt ($ $values (arg)))
-            (exclude-var functions arg))
-           (($ $primcall name args)
-            (exclude-vars functions args))
-           (($ $prompt escape? tag handler)
-            (exclude-var functions tag))))
+           (($ $primcall name param args)
+            (exclude-vars functions args))))
+        (($ $kargs _ _ ($ $branch kf kt src op param args))
+         (exclude-vars functions args))
+        (($ $kargs _ _ ($ $prompt k kh src escape? tag))
+         (exclude-var functions tag))
+        (($ $kargs _ _ ($ $throw src op param args))
+         (exclude-vars functions args))
         (_ functions)))
     (intmap-fold visit-cont conts functions)))
 
@@ -395,72 +374,119 @@ function set."
             (if (arity-matches? arity nargs)
                 body
                 (lp alt))))))))
-  (define (continue k src exp)
+  (define (inline-return cps k* kargs src nreq rest vals)
+    (define (build-list cps k src vals)
+      (match vals
+        (()
+         (with-cps cps
+           (build-term ($continue k src ($const '())))))
+        ((v . vals)
+         (with-cps cps
+           (letv pair tail)
+           (letk kdone ($kargs () () ($continue k src ($values (pair)))))
+           (letk ktail
+                 ($kargs () ()
+                   ($continue kdone src
+                     ($primcall 'scm-set!/immediate '(pair . 1) (pair tail)))))
+           (letk khead
+                 ($kargs ('pair) (pair)
+                   ($continue ktail src
+                     ($primcall 'scm-set!/immediate '(pair . 0) (pair v)))))
+           (letk ktail
+                 ($kargs ('tail) (tail)
+                   ($continue khead src
+                     ($primcall 'allocate-words/immediate '(pair . 2) ()))))
+           ($ (build-list ktail src vals))))))
+    (cond
+     ((and (not rest) (eqv? (length vals) nreq))
+      (with-cps cps
+        (build-term ($continue kargs src ($values vals)))))
+     ((and rest (<= nreq (length vals)))
+      (with-cps cps
+        (letv rest)
+        (letk krest ($kargs ('rest) (rest)
+                      ($continue kargs src
+                        ($values ,(append (list-head vals nreq)
+                                          (list rest))))))
+        ($ (build-list krest src (list-tail vals nreq)))))
+     (else
+      ;; Fallback case if values don't match.
+      (with-cps cps
+        (letv prim)
+        (letk kprim ($kargs ('prim) (prim)
+                      ($continue k* src ($call prim vals))))
+        (build-term ($continue kprim src ($prim 'values)))))))
+  (define (continue cps k src exp)
     (define (lookup-return-cont k)
       (match (return-subst k)
         (#f k)
         (k (lookup-return-cont k))))
     (let ((k* (lookup-return-cont k)))
       (if (eq? k k*)
-          (build-term ($continue k src ,exp))
-          ;; We are contifying this return.  It must be a call, a
-          ;; $values expression, or a return primcall.  k* will be
-          ;; either a $ktail or a $kreceive continuation.  CPS has this
-          ;; thing though where $kreceive can't be the target of a
-          ;; $values expression, and "return" can only continue to a
-          ;; tail continuation, so we might have to rewrite to a
-          ;; "values" primcall.
-          (build-term
-            ($continue k* src
-              ,(match (intmap-ref conts k*)
-                 (($ $kreceive)
-                  (match exp
-                    (($ $call) exp)
-                    ;; A primcall that can continue to $ktail can also
-                    ;; continue to $kreceive.
-                    (($ $primcall) exp)
-                    (($ $values vals)
-                     (build-exp ($primcall 'values vals)))))
-                 (($ $ktail) exp)))))))
-  (define (visit-exp k src exp)
+          (with-cps cps (build-term ($continue k src ,exp)))
+          ;; We are contifying this return.  It must be a call or a
+          ;; $values expression.  k* will be either a $ktail or a
+          ;; $kreceive continuation.
+          (match (intmap-ref conts k*)
+            (($ $kreceive ($ $arity req () rest () #f) kargs)
+             (match exp
+               (($ $call)
+                (with-cps cps (build-term ($continue k* src ,exp))))
+               ;; We need to punch through the $kreceive; otherwise we'd
+               ;; have to rewrite as a call to the 'values primitive.
+               (($ $values vals)
+                (inline-return cps k* kargs src (length req) rest vals))))
+            (($ $ktail)
+             (with-cps cps (build-term ($continue k* src ,exp))))))))
+  (define (visit-exp cps k src exp)
     (match exp
       (($ $call proc args)
        ;; If proc is contifiable, replace call with jump.
        (match (call-subst proc)
-         (#f (continue k src exp))
+         (#f (continue cps k src exp))
          (kfun
           (let ((body (find-body kfun (length args))))
-            (build-term ($continue body src ($values args)))))))
+            (with-cps cps
+              (build-term ($continue body src ($values args))))))))
       (($ $fun kfun)
        ;; If the function's tail continuation has been
        ;; substituted, that means it has been contified.
        (if (return-subst (tail-label conts kfun))
-           (continue k src (build-exp ($values ())))
-           (continue k src exp)))
+           (continue cps k src (build-exp ($values ())))
+           (continue cps k src exp)))
       (($ $rec names vars funs)
        (match (filter (match-lambda ((n v f) (not (call-subst v))))
                       (map list names vars funs))
-         (() (continue k src (build-exp ($values ()))))
+         (() (continue cps k src (build-exp ($values ()))))
          (((names vars funs) ...)
-          (continue k src (build-exp ($rec names vars funs))))))
-      (_ (continue k src exp))))
+          (continue cps k src (build-exp ($rec names vars funs))))))
+      (_ (continue cps k src exp))))
+  (define (visit-term cps term)
+    (match term
+      (($ $continue k src exp)
+       (visit-exp cps k src exp))
+      ((or ($ $branch) ($ $prompt) ($ $throw))
+       (with-cps cps term))))
 
   ;; Renumbering is not strictly necessary but some passes may not be
   ;; equipped to deal with stale $kfun nodes whose bodies have been
   ;; wired into other functions.
   (renumber
-   (intmap-map
-    (lambda (label cont)
-      (match cont
-        (($ $kargs names vars ($ $continue k src exp))
-         ;; Remove bindings for functions that have been contified.
-         (match (filter (match-lambda ((name var) (not (call-subst var))))
-                        (map list names vars))
-           (((names vars) ...)
-            (build-cont
-              ($kargs names vars ,(visit-exp k src exp))))))
-        (_ cont)))
-    conts)))
+   (with-fresh-name-state conts
+     (intmap-fold
+      (lambda (label cont out)
+        (match cont
+          (($ $kargs names vars term)
+           ;; Remove bindings for functions that have been contified.
+           (match (filter (match-lambda ((name var) (not (call-subst var))))
+                          (map list names vars))
+             (((names vars) ...)
+              (with-cps out
+                (let$ term (visit-term term))
+                (setk label ($kargs names vars ,term))))))
+          (_ out)))
+      conts
+      conts))))
 
 (define (contify conts)
   ;; FIXME: Renumbering isn't really needed but dead continuations may
