@@ -1,6 +1,6 @@
 ;;; Effects analysis on CPS
 
-;; Copyright (C) 2011, 2012, 2013, 2014, 2015 Free Software Foundation, Inc.
+;; Copyright (C) 2011-2015, 2017, 2018 Free Software Foundation, Inc.
 
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -42,6 +42,7 @@
 (define-module (language cps effects-analysis)
   #:use-module (language cps)
   #:use-module (language cps utils)
+  #:use-module (language cps intset)
   #:use-module (language cps intmap)
   #:use-module (ice-9 match)
   #:export (expression-effects
@@ -55,8 +56,6 @@
 
             &fluid
             &prompt
-            &car
-            &cdr
             &vector
             &box
             &module
@@ -78,12 +77,10 @@
             &no-effects
             &all-effects
 
-            exclude-effects
-            effect-free?
-            constant?
             causes-effect?
             causes-all-effects?
-            effect-clobbers?))
+            effect-clobbers?
+            compute-clobber-map))
 
 (define-syntax define-flags
   (lambda (x)
@@ -188,7 +185,13 @@
   &bytevector
 
   ;; Indicates a dependency on a free variable of a closure.
-  &closure)
+  &closure
+
+  ;; Indicates a dependency on a raw bitmask, measured in 32-bit units.
+  &bitmask
+
+  ;; Indicates a dependency on the value of a cache cell.
+  &cache)
 
 (define-inlinable (&field kind field)
   (ash (logior (ash field &memory-kind-bits) kind) &effect-kind-bits))
@@ -211,9 +214,6 @@
   (identifier-syntax
    (logior &all-effect-kinds (&object &unknown-memory-kinds))))
 
-(define-inlinable (constant? effects)
-  (zero? effects))
-
 (define-inlinable (causes-effect? x effects)
   (not (zero? (logand x effects))))
 
@@ -236,51 +236,93 @@ is or might be a read or a write to the same location as A."
        (not (zero? (logand b (logior &read &write))))
        (locations-same?)))
 
-(define-inlinable (indexed-field kind var constants)
-  (let ((val (intmap-ref constants var (lambda (_) #f))))
-    (if (and (exact-integer? val) (<= 0 val))
-        (&field kind val)
-        (&object kind))))
+(define (compute-clobber-map effects)
+  "For the map LABEL->EFFECTS, compute a map LABEL->LABELS indicating
+the LABELS that are clobbered by the effects of LABEL."
+  (let ((clobbered-by-write (make-hash-table)))
+    (intmap-fold
+     (lambda (label fx)
+       ;; Unless an expression causes a read, it isn't clobbered by
+       ;; anything.
+       (when (causes-effect? fx &read)
+         (let ((me (intset label)))
+           (define (add! kind field)
+             (let* ((k (logior (ash field &memory-kind-bits) kind))
+                    (clobber (hashv-ref clobbered-by-write k empty-intset)))
+               (hashv-set! clobbered-by-write k (intset-union me clobber))))
+           ;; Clobbered by write to specific field of this memory
+           ;; kind, write to any field of this memory kind, or
+           ;; write to any field of unknown memory kinds.
+           (let* ((loc (ash fx (- &effect-kind-bits)))
+                  (kind (logand loc &memory-kind-mask))
+                  (field (ash loc (- &memory-kind-bits))))
+             (add! kind field)
+             (add! kind -1)
+             (add! &unknown-memory-kinds -1))))
+       (values))
+     effects)
+    (intmap-map (lambda (label fx)
+                  (if (causes-effect? fx &write)
+                      (hashv-ref clobbered-by-write
+                                 (ash fx (- &effect-kind-bits))
+                                 empty-intset)
+                      empty-intset))
+                effects)))
 
 (define *primitive-effects* (make-hash-table))
 
-(define-syntax-rule (define-primitive-effects* constants
+(define-syntax-rule (define-primitive-effects* param
                       ((name . args) effects ...)
                       ...)
   (begin
     (hashq-set! *primitive-effects* 'name
                 (case-lambda*
-                 ((constants . args) (logior effects ...))
+                 ((param . args) (logior effects ...))
                  (_ &all-effects)))
     ...))
 
 (define-syntax-rule (define-primitive-effects ((name . args) effects ...) ...)
-  (define-primitive-effects* constants ((name . args) effects ...) ...))
+  (define-primitive-effects* param ((name . args) effects ...) ...))
 
 ;; Miscellaneous.
 (define-primitive-effects
+  ((load-const/unlikely))
   ((values . _)))
 
 ;; Generic effect-free predicates.
 (define-primitive-effects
-  ((eq? . _))
-  ((eqv? . _))
-  ((equal? . _))
-  ((pair? arg))
+  ((eq? x y))
+  ((equal? x y))
+  ((fixnum? arg))
+  ((char? arg))
+  ((eq-null? arg))
+  ((eq-nil? arg))
+  ((eq-false? arg))
+  ((eq-true? arg))
+  ((unspecified? arg))
+  ((undefined? arg))
+  ((eof-object? arg))
   ((null? arg))
-  ((nil? arg ))
+  ((false? arg))
+  ((nil? arg))
+  ((heap-object? arg))
+  ((pair? arg))
   ((symbol? arg))
   ((variable? arg))
   ((vector? arg))
   ((struct? arg))
   ((string? arg))
   ((number? arg))
-  ((char? arg))
   ((bytevector? arg))
   ((keyword? arg))
   ((bitvector? arg))
   ((procedure? arg))
-  ((thunk? arg)))
+  ((thunk? arg))
+  ((heap-number? arg))
+  ((bignum? arg))
+  ((flonum? arg))
+  ((compnum? arg))
+  ((fracnum? arg)))
 
 ;; Fluids.
 (define-primitive-effects
@@ -301,144 +343,152 @@ is or might be a read or a write to the same location as A."
 (define-primitive-effects
   ((make-prompt-tag #:optional arg) (&allocate &unknown-memory-kinds)))
 
-;; Pairs.
-(define-primitive-effects
-  ((cons a b)                      (&allocate &pair))
-  ((list . _)                      (&allocate &pair))
-  ((car x)                         (&read-field &pair 0)       &type-check)
-  ((set-car! x y)                  (&write-field &pair 0)      &type-check)
-  ((cdr x)                         (&read-field &pair 1)       &type-check)
-  ((set-cdr! x y)                  (&write-field &pair 1)      &type-check)
-  ((memq x y)                      (&read-object &pair)        &type-check)
-  ((memv x y)                      (&read-object &pair)        &type-check)
-  ((list? arg)                     (&read-field &pair 1))
-  ((length l)                      (&read-field &pair 1)       &type-check))
+;; Generic objects.
+(define (annotation->memory-kind annotation)
+  (match annotation
+    ('pair &pair)
+    ('vector &vector)
+    ('string &string)
+    ('stringbuf &string)
+    ('bytevector &bytevector)
+    ('bitmask &bitmask)
+    ('box &box)
+    ('closure &closure)
+    ('struct &struct)
+    ('atomic-box &unknown-memory-kinds)))
 
-;; Variables.
-(define-primitive-effects
-  ((box v)                         (&allocate &box))
-  ((box-ref v)                     (&read-object &box)         &type-check)
-  ((box-set! v x)                  (&write-object &box)        &type-check))
-
-;; Vectors.
-(define (vector-field n constants)
-  (indexed-field &vector n constants))
-(define (read-vector-field n constants)
-  (logior &read (vector-field n constants)))
-(define (write-vector-field n constants)
-  (logior &write (vector-field n constants)))
-(define-primitive-effects* constants
-  ((vector . _)                    (&allocate &vector))
-  ((make-vector n init)            (&allocate &vector))
-  ((make-vector/immediate n init)  (&allocate &vector))
-  ((vector-ref v n)                (read-vector-field n constants) &type-check)
-  ((vector-ref/immediate v n)      (read-vector-field n constants) &type-check)
-  ((vector-set! v n x)             (write-vector-field n constants) &type-check)
-  ((vector-set!/immediate v n x)   (write-vector-field n constants) &type-check)
-  ((vector-length v)                                           &type-check))
-
-;; Structs.
-(define (struct-field n constants)
-  (indexed-field &struct n constants))
-(define (read-struct-field n constants)
-  (logior &read (struct-field n constants)))
-(define (write-struct-field n constants)
-  (logior &write (struct-field n constants)))
-(define-primitive-effects* constants
-  ((allocate-struct vt n)          (&allocate &struct)         &type-check)
-  ((allocate-struct/immediate v n) (&allocate &struct)         &type-check)
-  ((make-struct vt ntail . _)      (&allocate &struct)         &type-check)
-  ((make-struct/no-tail vt . _)    (&allocate &struct)         &type-check)
-  ((struct-ref s n)                (read-struct-field n constants) &type-check)
-  ((struct-ref/immediate s n)      (read-struct-field n constants) &type-check)
-  ((struct-set! s n x)             (write-struct-field n constants) &type-check)
-  ((struct-set!/immediate s n x)   (write-struct-field n constants) &type-check)
-  ((struct-vtable s)                                           &type-check))
+(define-primitive-effects* param
+  ((allocate-words size)           (&allocate (annotation->memory-kind param)))
+  ((allocate-words/immediate)      (match param
+                                     ((ann . size)
+                                      (&allocate
+                                       (annotation->memory-kind ann)))))
+  ((scm-ref obj idx)               (&read-object
+                                    (annotation->memory-kind param)))
+  ((scm-ref/tag obj)               (&read-field
+                                    (annotation->memory-kind param) 0))
+  ((scm-ref/immediate obj)         (match param
+                                     ((ann . idx)
+                                      (&read-field
+                                       (annotation->memory-kind ann) idx))))
+  ((scm-set! obj idx val)          (&write-object
+                                    (annotation->memory-kind param)))
+  ((scm-set/tag! obj val)          (&write-field
+                                    (annotation->memory-kind param) 0))
+  ((scm-set!/immediate obj val)    (match param
+                                     ((ann . idx)
+                                      (&write-field
+                                       (annotation->memory-kind ann) idx))))
+  ((word-ref obj idx)              (&read-object
+                                    (annotation->memory-kind param)))
+  ((word-ref/immediate obj)        (match param
+                                     ((ann . idx)
+                                      (&read-field
+                                       (annotation->memory-kind ann) idx))))
+  ((word-set! obj idx val)         (&read-object
+                                    (annotation->memory-kind param)))
+  ((word-set!/immediate obj val)   (match param
+                                     ((ann . idx)
+                                      (&write-field
+                                       (annotation->memory-kind ann) idx))))
+  ((pointer-ref/immediate obj)     (match param
+                                     ((ann . idx)
+                                      (&read-field
+                                       (annotation->memory-kind ann) idx))))
+  ((pointer-set!/immediate obj val)
+                                   (match param
+                                     ((ann . idx)
+                                      (&write-field
+                                       (annotation->memory-kind ann) idx))))
+  ((tail-pointer-ref/immediate obj)))
 
 ;; Strings.
 (define-primitive-effects
-  ((string-ref s n)                (&read-object &string)      &type-check)
   ((string-set! s n c)             (&write-object &string)     &type-check)
   ((number->string _)              (&allocate &string)         &type-check)
-  ((string->number _)              (&read-object &string)      &type-check)
-  ((string-length s)                                           &type-check))
+  ((string->number _)              (&read-object &string)      &type-check))
 
 ;; Unboxed floats and integers.
 (define-primitive-effects
   ((scm->f64 _)                                                &type-check)
-  ((load-f64 _))
+  ((load-f64))
   ((f64->scm _))
   ((scm->u64 _)                                                &type-check)
   ((scm->u64/truncate _)                                       &type-check)
-  ((load-u64 _))
+  ((load-u64))
   ((u64->scm _))
+  ((u64->scm/unlikely _))
   ((scm->s64 _)                                                &type-check)
-  ((load-s64 _))
-  ((s64->scm _)))
+  ((load-s64))
+  ((s64->scm _))
+  ((s64->scm/unlikely _))
+  ((u64->s64 _))
+  ((s64->u64 _))
+  ((assume-u64 _))
+  ((assume-s64 _))
+  ((untag-fixnum _))
+  ((tag-fixnum _))
+  ((tag-fixnum/unlikely _)))
 
-;; Bytevectors.
-(define-primitive-effects
-  ((bv-length _)                                               &type-check)
+;; Pointers.
+(define-primitive-effects* param
+  ((u8-ref obj bv n)               (&read-object (annotation->memory-kind param)))
+  ((s8-ref obj bv n)               (&read-object (annotation->memory-kind param)))
+  ((u16-ref obj bv n)              (&read-object (annotation->memory-kind param)))
+  ((s16-ref obj bv n)              (&read-object (annotation->memory-kind param)))
+  ((u32-ref obj bv n)              (&read-object (annotation->memory-kind param)))
+  ((s32-ref obj bv n)              (&read-object (annotation->memory-kind param)))
+  ((u64-ref obj bv n)              (&read-object (annotation->memory-kind param)))
+  ((s64-ref obj bv n)              (&read-object (annotation->memory-kind param)))
+  ((f32-ref obj bv n)              (&read-object (annotation->memory-kind param)))
+  ((f64-ref obj bv n)              (&read-object (annotation->memory-kind param)))
 
-  ((bv-u8-ref bv n)                (&read-object &bytevector)  &type-check)
-  ((bv-s8-ref bv n)                (&read-object &bytevector)  &type-check)
-  ((bv-u16-ref bv n)               (&read-object &bytevector)  &type-check)
-  ((bv-s16-ref bv n)               (&read-object &bytevector)  &type-check)
-  ((bv-u32-ref bv n)               (&read-object &bytevector)  &type-check)
-  ((bv-s32-ref bv n)               (&read-object &bytevector)  &type-check)
-  ((bv-u64-ref bv n)               (&read-object &bytevector)  &type-check)
-  ((bv-s64-ref bv n)               (&read-object &bytevector)  &type-check)
-  ((bv-f32-ref bv n)               (&read-object &bytevector)  &type-check)
-  ((bv-f64-ref bv n)               (&read-object &bytevector)  &type-check)
-
-  ((bv-u8-set! bv n x)             (&write-object &bytevector) &type-check)
-  ((bv-s8-set! bv n x)             (&write-object &bytevector) &type-check)
-  ((bv-u16-set! bv n x)            (&write-object &bytevector) &type-check)
-  ((bv-s16-set! bv n x)            (&write-object &bytevector) &type-check)
-  ((bv-u32-set! bv n x)            (&write-object &bytevector) &type-check)
-  ((bv-s32-set! bv n x)            (&write-object &bytevector) &type-check)
-  ((bv-u64-set! bv n x)            (&write-object &bytevector) &type-check)
-  ((bv-s64-set! bv n x)            (&write-object &bytevector) &type-check)
-  ((bv-f32-set! bv n x)            (&write-object &bytevector) &type-check)
-  ((bv-f64-set! bv n x)            (&write-object &bytevector) &type-check))
-
-;; Closures.
-(define (closure-field n constants)
-  (indexed-field &closure n constants))
-(define (read-closure-field n constants)
-  (logior &read (closure-field n constants)))
-(define (write-closure-field n constants)
-  (logior &write (closure-field n constants)))
-(define-primitive-effects* constants
-  ((free-ref closure idx)          (read-closure-field idx constants))
-  ((free-set! closure idx val)     (write-closure-field idx constants)))
+  ((u8-set! obj bv n x)            (&write-object (annotation->memory-kind param)))
+  ((s8-set! obj bv n x)            (&write-object (annotation->memory-kind param)))
+  ((u16-set! obj bv n x)           (&write-object (annotation->memory-kind param)))
+  ((s16-set! obj bv n x)           (&write-object (annotation->memory-kind param)))
+  ((u32-set! obj bv n x)           (&write-object (annotation->memory-kind param)))
+  ((s32-set! obj bv n x)           (&write-object (annotation->memory-kind param)))
+  ((u64-set! obj bv n x)           (&write-object (annotation->memory-kind param)))
+  ((s64-set! obj bv n x)           (&write-object (annotation->memory-kind param)))
+  ((f32-set! obj bv n x)           (&write-object (annotation->memory-kind param)))
+  ((f64-set! obj bv n x)           (&write-object (annotation->memory-kind param))))
 
 ;; Modules.
 (define-primitive-effects
   ((current-module)                (&read-object &module))
-  ((cache-current-module! m scope) (&write-object &box))
-  ((resolve name bound?)           (&read-object &module)      &type-check)
-  ((cached-toplevel-box scope name bound?)                     &type-check)
-  ((cached-module-box mod name public? bound?)                 &type-check)
-  ((define! name)                  (&read-object &module)))
+  ((cache-current-module! m)       (&write-object &cache))
+  ((resolve name)                  (&read-object &module)      &type-check)
+  ((resolve-module mod)            (&read-object &module)      &type-check)
+  ((lookup mod name)               (&read-object &module)      &type-check)
+  ((cached-toplevel-box)                                       &type-check)
+  ((cached-module-box)                                         &type-check)
+  ((define! mod name)              (&read-object &module)))
+
+;; Cache cells.
+(define-primitive-effects
+  ((cache-ref)                     (&read-object &cache))
+  ((cache-set! x)                  (&write-object &cache)))
 
 ;; Numbers.
 (define-primitive-effects
+  ((heap-numbers-equal? . _))
   ((= . _)                         &type-check)
+  ((<= . _)                         &type-check)
   ((< . _)                         &type-check)
-  ((> . _)                         &type-check)
-  ((<= . _)                        &type-check)
-  ((>= . _)                        &type-check)
   ((u64-= . _))
+  ((u64-imm-= . _))
   ((u64-< . _))
-  ((u64-> . _))
-  ((u64-<= . _))
-  ((u64->= . _))
-  ((u64-<-scm . _)                 &type-check)
-  ((u64-<=-scm . _)                &type-check)
-  ((u64-=-scm . _)                 &type-check)
-  ((u64->=-scm . _)                 &type-check)
-  ((u64->-scm . _)                 &type-check)
+  ((u64-imm-< . _))
+  ((imm-u64-< . _))
+  ((s64-= . _))
+  ((s64-imm-= . _))
+  ((s64-< . _))
+  ((s64-imm-< . _))
+  ((imm-s64-< . _))
+  ((f64-= . _))
+  ((f64-< . _))
+  ((f64-<= . _))
   ((zero? . _)                     &type-check)
   ((add . _)                       &type-check)
   ((add/immediate . _)             &type-check)
@@ -456,6 +506,12 @@ is or might be a read or a write to the same location as A."
   ((uadd/immediate . _))
   ((usub/immediate . _))
   ((umul/immediate . _))
+  ((sadd . _))
+  ((ssub . _))
+  ((smul . _))
+  ((sadd/immediate . _))
+  ((ssub/immediate . _))
+  ((smul/immediate . _))
   ((quo . _)                       &type-check)
   ((rem . _)                       &type-check)
   ((mod . _)                       &type-check)
@@ -469,7 +525,10 @@ is or might be a read or a write to the same location as A."
   ((inexact? _)                    &type-check)
   ((even? _)                       &type-check)
   ((odd? _)                        &type-check)
-  ((ash n m)                       &type-check)
+  ((rsh n m)                       &type-check)
+  ((lsh n m)                       &type-check)
+  ((rsh/immediate n)               &type-check)
+  ((lsh/immediate n)               &type-check)
   ((logand . _)                    &type-check)
   ((logior . _)                    &type-check)
   ((logxor . _)                    &type-check)
@@ -480,9 +539,13 @@ is or might be a read or a write to the same location as A."
   ((ulogxor . _))
   ((ulogsub . _))
   ((ursh . _))
+  ((srsh . _))
   ((ulsh . _))
+  ((slsh . _))
   ((ursh/immediate . _))
+  ((srsh/immediate . _))
   ((ulsh/immediate . _))
+  ((slsh/immediate . _))
   ((logtest a b)                   &type-check)
   ((logbit? a b)                   &type-check)
   ((sqrt _)                        &type-check)
@@ -490,52 +553,55 @@ is or might be a read or a write to the same location as A."
 
 ;; Characters.
 (define-primitive-effects
-  ((integer->char _)               &type-check)
-  ((char->integer _)               &type-check))
+  ((untag-char _))
+  ((tag-char _)))
 
 ;; Atomics are a memory and a compiler barrier; they cause all effects
 ;; so no need to have a case for them here.  (Though, see
 ;; https://jfbastien.github.io/no-sane-compiler/.)
 
-(define (primitive-effects constants name args)
+(define (primitive-effects param name args)
   (let ((proc (hashq-ref *primitive-effects* name)))
     (if proc
-        (apply proc constants args)
+        (apply proc param args)
         &all-effects)))
 
-(define (expression-effects exp constants)
+(define (expression-effects exp)
   (match exp
-    ((or ($ $const) ($ $prim) ($ $values))
+    ((or ($ $const) ($ $prim) ($ $values) ($ $code) ($ $const-fun))
      &no-effects)
-    (($ $closure _ 0)
-     &no-effects)
-    ((or ($ $fun) ($ $rec) ($ $closure))
+    ((or ($ $fun) ($ $rec))
      (&allocate &unknown-memory-kinds))
-    (($ $prompt)
-     (&write-object &prompt))
     ((or ($ $call) ($ $callk))
      &all-effects)
-    (($ $branch k exp)
-     (expression-effects exp constants))
-    (($ $primcall name args)
-     (primitive-effects constants name args))))
+    (($ $primcall name param args)
+     (primitive-effects param name args))))
 
 (define (compute-effects conts)
-  (let ((constants (compute-constant-values conts)))
-    (intmap-map
-     (lambda (label cont)
-       (match cont
-         (($ $kargs names syms ($ $continue k src exp))
-          (expression-effects exp constants))
-         (($ $kreceive arity kargs)
-          (match arity
-            (($ $arity _ () #f () #f) &type-check)
-            (($ $arity () () _ () #f) (&allocate &pair))
-            (($ $arity _ () _ () #f) (logior (&allocate &pair) &type-check))))
-         (($ $kfun) &type-check)
-         (($ $kclause) &type-check)
-         (($ $ktail) &no-effects)))
-     conts)))
+  (intmap-map
+   (lambda (label cont)
+     (match cont
+       (($ $kargs names syms ($ $continue k src exp))
+        (expression-effects exp))
+       (($ $kargs names syms ($ $branch kf kt src op param args))
+        (primitive-effects param op args))
+       (($ $kargs names syms ($ $prompt))
+        ;; Although the "main" path just writes &prompt, we don't know
+        ;; what nonlocal predecessors of the handler do, so we
+        ;; conservatively assume &all-effects.
+        &all-effects)
+       (($ $kargs names syms ($ $throw))
+        ;; A reachable "throw" term can never be elided.
+        &all-effects)
+       (($ $kreceive arity kargs)
+        (match arity
+          (($ $arity _ () #f () #f) &type-check)
+          (($ $arity () () _ () #f) (&allocate &pair))
+          (($ $arity _ () _ () #f) (logior (&allocate &pair) &type-check))))
+       (($ $kfun) &type-check)
+       (($ $kclause) &type-check)
+       (($ $ktail) &no-effects)))
+   conts))
 
 ;; There is a way to abuse effects analysis in CSE to also do scalar
 ;; replacement, effectively adding `car' and `cdr' expressions to `cons'
