@@ -1,5 +1,5 @@
 ;;; Functional name maps
-;;; Copyright (C) 2014, 2015 Free Software Foundation, Inc.
+;;; Copyright (C) 2014, 2015, 2017 Free Software Foundation, Inc.
 ;;;
 ;;; This library is free software: you can redistribute it and/or modify
 ;;; it under the terms of the GNU Lesser General Public License as
@@ -33,8 +33,8 @@
 (define-module (language cps intmap)
   #:use-module (srfi srfi-9)
   #:use-module (srfi srfi-9 gnu)
-  #:use-module (srfi srfi-18)
   #:use-module (ice-9 match)
+  #:use-module ((ice-9 threads) #:select (current-thread))
   #:export (empty-intmap
             intmap?
             transient-intmap?
@@ -386,7 +386,9 @@ already, and always calls the meet procedure."
        (let ((root* (remove (- i min) shift root)))
          (if (eq? root root*)
              map
-             (make-intmap/prune min shift root*))))
+             (if (absent? root*)
+                 empty-intmap
+                 (make-intmap/prune min shift root*)))))
       (else map)))
     (($ <transient-intmap>)
      (intmap-remove (persistent-intmap map) i))))
@@ -511,6 +513,8 @@ already, and always calls the meet procedure."
 
 (define intmap-fold
   (case-lambda
+    ((f map)
+     ((make-intmap-folder #t) f map))
     ((f map seed)
      ((make-intmap-folder #t seed) f map seed))
     ((f map seed0 seed1)
@@ -520,6 +524,8 @@ already, and always calls the meet procedure."
 
 (define intmap-fold-right
   (case-lambda
+    ((f map)
+     ((make-intmap-folder #f) f map))
     ((f map seed)
      ((make-intmap-folder #f seed) f map seed))
     ((f map seed0 seed1)
