@@ -1,6 +1,6 @@
 ;; Continuation-passing style (CPS) intermediate language (IL)
 
-;; Copyright (C) 2013, 2014, 2015 Free Software Foundation, Inc.
+;; Copyright (C) 2013, 2014, 2015, 2017, 2018 Free Software Foundation, Inc.
 
 ;;;; This library is free software; you can redistribute it and/or
 ;;;; modify it under the terms of the GNU Lesser General Public
@@ -37,16 +37,13 @@
             lookup-slot
             lookup-maybe-slot
             lookup-representation
-            lookup-constant-value
-            lookup-maybe-constant-value
             lookup-nlocals
             lookup-call-proc-slot
             lookup-parallel-moves
             lookup-slot-map))
 
 (define-record-type $allocation
-  (make-allocation slots representations constant-values call-allocs
-                   shuffles frame-size)
+  (make-allocation slots representations call-allocs shuffles frame-size)
   allocation?
 
   ;; A map of VAR to slot allocation.  A slot allocation is an integer,
@@ -59,12 +56,8 @@
   ;;
   (representations allocation-representations)
 
-  ;; A map of VAR to constant value, for variables with constant values.
-  ;;
-  (constant-values allocation-constant-values)
-
   ;; A map of LABEL to /call allocs/, for expressions that continue to
-  ;; $kreceive continuations: non-tail calls and $prompt expressions.
+  ;; $kreceive continuations: non-tail calls and $prompt terms.
   ;;
   ;; A call alloc contains two pieces of information: the call's /proc
   ;; slot/ and a /dead slot map/.  The proc slot indicates the slot of a
@@ -111,20 +104,6 @@
 
 (define *absent* (list 'absent))
 
-(define (lookup-constant-value var allocation)
-  (let ((value (intmap-ref (allocation-constant-values allocation) var
-                           (lambda (_) *absent*))))
-    (when (eq? value *absent*)
-      (error "Variable does not have constant value" var))
-    value))
-
-(define (lookup-maybe-constant-value var allocation)
-  (let ((value (intmap-ref (allocation-constant-values allocation) var
-                           (lambda (_) *absent*))))
-    (if (eq? value *absent*)
-        (values #f #f)
-        (values #t value))))
-
 (define (lookup-call-alloc k allocation)
   (intmap-ref (allocation-call-allocs allocation) k))
 
@@ -167,22 +146,22 @@ by a label, respectively."
          (return (intset self) empty-intset))
         (($ $kargs _ _ ($ $continue k src exp))
          (match exp
-           ((or ($ $const) ($ $closure))
+           ((or ($ $const) ($ $const-fun) ($ $code))
             (return (get-defs k) empty-intset))
            (($ $call proc args)
             (return (get-defs k) (intset-add (vars->intset args) proc)))
            (($ $callk _ proc args)
             (return (get-defs k) (intset-add (vars->intset args) proc)))
-           (($ $primcall name args)
+           (($ $primcall name param args)
             (return (get-defs k) (vars->intset args)))
-           (($ $branch kt ($ $primcall name args))
-            (return empty-intset (vars->intset args)))
-           (($ $branch kt ($ $values args))
-            (return empty-intset (vars->intset args)))
            (($ $values args)
-            (return (get-defs k) (vars->intset args)))
-           (($ $prompt escape? tag handler)
-            (return empty-intset (intset tag)))))
+            (return (get-defs k) (vars->intset args)))))
+        (($ $kargs _ _ ($ $branch kf kt src op param args))
+         (return empty-intset (vars->intset args)))
+        (($ $kargs _ _ ($ $prompt k kh src escape? tag))
+         (return empty-intset (intset tag)))
+        (($ $kargs _ _ ($ $throw src op param args))
+         (return empty-intset (vars->intset args)))
         (($ $kclause arity body alt)
          (return (get-defs body) empty-intset))
         (($ $kreceive arity kargs)
@@ -245,26 +224,20 @@ body continuation in the prompt."
         ((zero? level) labels)
         ((intset-ref labels label) labels)
         (else
-         (match (intmap-ref conts label)
-           (($ $ktail)
-            ;; Possible for bailouts; never reached and not part of
-            ;; prompt body.
-            labels)
-           (cont
-            (let ((labels (intset-add! labels label)))
-              (match cont
-                (($ $kreceive arity k) (visit-cont k level labels))
-                (($ $kargs names syms ($ $continue k src ($ $primcall 'wind)))
-                 (visit-cont k (1+ level) labels))
-                (($ $kargs names syms
-                           ($ $continue k src ($ $prompt escape? tag handler)))
-                 (visit-cont handler level (visit-cont k (1+ level) labels)))
-                (($ $kargs names syms ($ $continue k src ($ $primcall 'unwind)))
-                 (visit-cont k (1- level) labels))
-                (($ $kargs names syms ($ $continue k src ($ $branch kt)))
-                 (visit-cont k level (visit-cont kt level labels)))
-                (($ $kargs names syms ($ $continue k src exp))
-                 (visit-cont k level labels)))))))))))
+         (let ((labels (intset-add! labels label)))
+           (match (intmap-ref conts label)
+             (($ $kreceive arity k) (visit-cont k level labels))
+             (($ $kargs names syms ($ $continue k src ($ $primcall 'wind)))
+              (visit-cont k (1+ level) labels))
+             (($ $kargs names syms ($ $continue k src ($ $primcall 'unwind)))
+              (visit-cont k (1- level) labels))
+             (($ $kargs names syms ($ $continue k src exp))
+              (visit-cont k level labels))
+             (($ $kargs names syms ($ $branch kf kt))
+              (visit-cont kf level (visit-cont kt level labels)))
+             (($ $kargs names syms ($ $prompt k kh src escape? tag))
+              (visit-cont kh level (visit-cont k (1+ level) labels)))
+             (($ $kargs names syms ($ $throw)) labels))))))))
   (define (visit-prompt label handler succs)
     (let ((body (compute-prompt-body label)))
       (define (out-or-back-edge? label)
@@ -273,10 +246,9 @@ body continuation in the prompt."
         ;; continuations that postdominate the rest of the body.  Unless
         ;; you pass #:complete? #t, we only invoke F on continuations
         ;; that can leave the body, or on back-edges in loops.
-        (intset-any (lambda (succ)
-                      (or (not (intset-ref body succ))
-                          (<= succ label)))
-                    (intmap-ref succs label)))
+        (not (intset-any (lambda (succ)
+                           (and (intset-ref body succ) (< label succ)))
+                         (intmap-ref succs label))))
       (intset-fold (lambda (pred succs)
                      (intmap-replace succs pred handler intset-add))
                    (if complete? body (intset-filter out-or-back-edge? body))
@@ -284,9 +256,8 @@ body continuation in the prompt."
   (intmap-fold
    (lambda (label cont succs)
      (match cont
-       (($ $kargs _ _
-           ($ $continue k _ ($ $prompt escape? tag handler)))
-        (visit-prompt k handler succs))
+       (($ $kargs _ _ ($ $prompt k kh))
+        (visit-prompt k kh succs))
        (_ succs)))
    conts
    succs))
@@ -344,50 +315,8 @@ the definitions that are live before and after LABEL, as intsets."
      (intset-union
       needs-slot
       (match cont
-        (($ $kargs _ _ ($ $continue k src exp))
-         (let ((defs (get-defs label)))
-           (define (defs+* uses)
-             (intset-union defs uses))
-           (define (defs+ use)
-             (intset-add defs use))
-           (match exp
-             (($ $const)
-              empty-intset)
-             (($ $primcall (or 'load-f64 'load-u64 'load-s64) (val))
-              empty-intset)
-             (($ $primcall 'free-ref (closure slot))
-              (defs+ closure))
-             (($ $primcall 'free-set! (closure slot value))
-              (defs+* (intset closure value)))
-             (($ $primcall 'cache-current-module! (mod . _))
-              (defs+ mod))
-             (($ $primcall 'cached-toplevel-box _)
-              defs)
-             (($ $primcall 'cached-module-box _)
-              defs)
-             (($ $primcall 'resolve (name bound?))
-              (defs+ name))
-             (($ $primcall 'make-vector/immediate (len init))
-              (defs+ init))
-             (($ $primcall 'vector-ref/immediate (v i))
-              (defs+ v))
-             (($ $primcall 'vector-set!/immediate (v i x))
-              (defs+* (intset v x)))
-             (($ $primcall 'allocate-struct/immediate (vtable nfields))
-              (defs+ vtable))
-             (($ $primcall 'struct-ref/immediate (s n))
-              (defs+ s))
-             (($ $primcall 'struct-set!/immediate (s n x))
-              (defs+* (intset s x)))
-             (($ $primcall (or 'add/immediate 'sub/immediate
-                               'uadd/immediate 'usub/immediate 'umul/immediate
-                               'ursh/immediate 'ulsh/immediate)
-                 (x y))
-              (defs+ x))
-             (($ $primcall 'builtin-ref (idx))
-              defs)
-             (_
-              (defs+* (get-uses label))))))
+        (($ $kargs)
+         (intset-union (get-defs label) (get-uses label)))
         (($ $kreceive arity k)
          ;; Only allocate results of function calls to slots if they are
          ;; used.
@@ -599,7 +528,7 @@ are comparable with eqv?.  A tmp slot may be used."
       (($ $kreceive arity kargs)
        (let* ((results (match (get-cont kargs)
                          (($ $kargs names vars) vars)))
-              (value-slots (integers (1+ proc-slot) (length results)))
+              (value-slots (integers proc-slot (length results)))
               (result-slots (get-slots results))
               ;; Filter out unused results.
               (value-slots (filter-map (lambda (val result) (and result val))
@@ -634,7 +563,7 @@ are comparable with eqv?.  A tmp slot may be used."
       (($ $ktail)
        (let* ((live (compute-live-slots label))
               (src-slots (get-slots args))
-              (dst-slots (integers 1 (length args)))
+              (dst-slots (integers 0 (length args)))
               (moves (parallel-move src-slots dst-slots
                                     (compute-tmp-slot live dst-slots))))
          (intmap-add! shuffles label moves)))
@@ -661,9 +590,9 @@ are comparable with eqv?.  A tmp slot may be used."
           (add-call-shuffles label k (cons proc args) shuffles))
          (($ $values args)
           (add-values-shuffles label k args shuffles))
-         (($ $prompt escape? tag handler)
-          (add-prompt-shuffles label k handler shuffles))
          (_ shuffles)))
+      (($ $kargs names vars ($ $prompt k kh src escape? tag))
+       (add-prompt-shuffles label k kh shuffles))
       (_ shuffles)))
 
   (persistent-intmap
@@ -694,14 +623,14 @@ are comparable with eqv?.  A tmp slot may be used."
                   (max (+ (get-proc-slot label) nargs) size)))
   (define (measure-cont label cont size)
     (match cont
-      (($ $kargs names vars ($ $continue k src exp))
+      (($ $kargs names vars term)
        (let ((size (max-size* vars size)))
-         (match exp
-           (($ $call proc args)
+         (match term
+           (($ $continue _ _ ($ $call proc args))
             (call-size label (1+ (length args)) size))
-           (($ $callk _ proc args)
+           (($ $continue _ _ ($ $callk _ proc args))
             (call-size label (1+ (length args)) size))
-           (($ $values args)
+           (($ $continue _ _ ($ $values args))
             (shuffle-size (get-shuffles label) size))
            (_ size))))
       (($ $kreceive)
@@ -776,7 +705,7 @@ are comparable with eqv?.  A tmp slot may be used."
   (define (allocate-values label k args slots)
     (match (intmap-ref cps k)
       (($ $ktail)
-       (allocate* args (integers 1 (length args))
+       (allocate* args (integers 0 (length args))
                   slots (compute-live-slots slots label)))
       (($ $kargs names vars)
        (allocate* args
@@ -818,21 +747,31 @@ are comparable with eqv?.  A tmp slot may be used."
               (intmap-add representations var
                           (intmap-ref representations arg)))
              (($ $primcall (or 'scm->f64 'load-f64
-                               'bv-f32-ref 'bv-f64-ref
+                               'f32-ref 'f64-ref
                                'fadd 'fsub 'fmul 'fdiv))
               (intmap-add representations var 'f64))
              (($ $primcall (or 'scm->u64 'scm->u64/truncate 'load-u64
-                               'char->integer
-                               'bv-length 'vector-length 'string-length
+                               's64->u64
+                               'assume-u64
                                'uadd 'usub 'umul
                                'ulogand 'ulogior 'ulogxor 'ulogsub 'ursh 'ulsh
                                'uadd/immediate 'usub/immediate 'umul/immediate
                                'ursh/immediate 'ulsh/immediate
-                               'bv-u8-ref 'bv-u16-ref 'bv-u32-ref 'bv-u64-ref))
+                               'u8-ref 'u16-ref 'u32-ref 'u64-ref
+                               'word-ref 'word-ref/immediate
+                               'untag-char))
               (intmap-add representations var 'u64))
-             (($ $primcall (or 'scm->s64 'load-s64
-                               'bv-s8-ref 'bv-s16-ref 'bv-s32-ref 'bv-s64-ref))
+             (($ $primcall (or 'untag-fixnum
+                               'assume-s64
+                               'scm->s64 'load-s64 'u64->s64
+                               'srsh 'srsh/immediate
+                               's8-ref 's16-ref 's32-ref 's64-ref))
               (intmap-add representations var 's64))
+             (($ $primcall (or 'pointer-ref/immediate
+                               'tail-pointer-ref/immediate))
+              (intmap-add representations var 'ptr))
+             (($ $code)
+              (intmap-add representations var 'u64))
              (_
               (intmap-add representations var 'scm))))
           (vars
@@ -842,6 +781,8 @@ are comparable with eqv?.  A tmp slot may be used."
                       (intmap-add representations var
                                   (intmap-ref representations arg)))
                     representations args vars))))))
+       (($ $kargs _ _ (or ($ $branch) ($ $prompt) ($ $throw)))
+        representations)
        (($ $kfun src meta self)
         (intmap-add representations self 'scm))
        (($ $kclause arity body alt)
@@ -860,23 +801,22 @@ are comparable with eqv?.  A tmp slot may be used."
   (let*-values (((defs uses) (compute-defs-and-uses cps))
                 ((representations) (compute-var-representations cps))
                 ((live-in live-out) (compute-live-variables cps defs uses))
-                ((constants) (compute-constant-values cps))
                 ((needs-slot) (compute-needs-slot cps defs uses))
                 ((lazy) (if precolor-calls?
                             (compute-lazy-vars cps live-in live-out defs
                                                needs-slot)
                             empty-intset)))
 
+    (define frame-size 3)
+
     (define (empty-live-slots)
       #b0)
 
     (define (compute-call-proc-slot live-slots)
-      (+ 2 (find-first-trailing-zero live-slots)))
+      (+ frame-size (find-first-trailing-zero live-slots)))
 
     (define (compute-prompt-handler-proc-slot live-slots)
-      (if (zero? live-slots)
-          0
-          (1- (find-first-trailing-zero live-slots))))
+      (find-first-trailing-zero live-slots))
 
     (define (get-cont label)
       (intmap-ref cps label))
@@ -916,7 +856,7 @@ are comparable with eqv?.  A tmp slot may be used."
            (#f slot-map)
            (slot
             (let ((desc (match (intmap-ref representations var)
-                          ((or 'u64 'f64 's64) slot-desc-live-raw)
+                          ((or 'u64 'f64 's64 'ptr) slot-desc-live-raw)
                           ('scm slot-desc-live-scm))))
               (logior slot-map (ash desc (* 2 slot)))))))
        live-vars 0))
@@ -983,11 +923,11 @@ are comparable with eqv?.  A tmp slot may be used."
                  (($ $kargs () ())
                   (values slots post-live))
                  (($ $kargs (_ . _) (_ . results))
-                  (let ((result-slots (integers (+ proc-slot 2)
+                  (let ((result-slots (integers (+ proc-slot 1)
                                                 (length results))))
                     (allocate* results result-slots slots post-live)))))
               ((slot-map) (compute-slot-map slots (intmap-ref live-out label)
-                                            (- proc-slot 2)))
+                                            (- proc-slot frame-size)))
               ((call) (make-call-alloc proc-slot slot-map)))
            (values slots
                    (intmap-add! call-allocs label call))))))
@@ -1022,10 +962,10 @@ are comparable with eqv?.  A tmp slot may be used."
              (((handler-live) (compute-live-in-slots slots handler))
               ((proc-slot) (compute-prompt-handler-proc-slot handler-live))
               ((slot-map)  (compute-slot-map slots (intmap-ref live-in handler)
-                                             (- proc-slot 2)))
+                                             (- proc-slot frame-size)))
               ((result-vars) (match (get-cont kargs)
                                (($ $kargs names vars) vars)))
-              ((value-slots) (integers (1+ proc-slot) (length result-vars)))
+              ((value-slots) (integers proc-slot (length result-vars)))
               ((slots result-live) (allocate* result-vars value-slots
                                               slots handler-live)))
            (values slots
@@ -1034,17 +974,17 @@ are comparable with eqv?.  A tmp slot may be used."
 
     (define (allocate-cont label cont slots call-allocs)
       (match cont
-        (($ $kargs names vars ($ $continue k src exp))
+        (($ $kargs names vars term)
          (let-values (((slots live) (allocate-defs label vars slots)))
-           (match exp
-             (($ $call proc args)
+           (match term
+             (($ $continue k src ($ $call proc args))
               (allocate-call label k (cons proc args) slots call-allocs live))
-             (($ $callk _ proc args)
+             (($ $continue k src ($ $callk _ proc args))
               (allocate-call label k (cons proc args) slots call-allocs live))
-             (($ $values args)
+             (($ $continue k src ($ $values args))
               (allocate-values label k args slots call-allocs))
-             (($ $prompt escape? tag handler)
-              (allocate-prompt label k handler slots call-allocs))
+             (($ $prompt k kh src escape? tag)
+              (allocate-prompt label k kh slots call-allocs))
              (_
               (values slots call-allocs)))))
         (_
@@ -1057,5 +997,4 @@ are comparable with eqv?.  A tmp slot may be used."
         (let* ((slots (allocate-lazy-vars cps slots calls live-in lazy))
                (shuffles (compute-shuffles cps slots calls live-in))
                (frame-size (compute-frame-size cps slots calls shuffles)))
-          (make-allocation slots representations constants calls
-                           shuffles frame-size))))))
+          (make-allocation slots representations calls shuffles frame-size))))))
